/*
 * VNCFramebuffer.m
 * A simple off-screen framebuffer optimized for drawing rectangles and raw images.
 *
 * Copyright (c) 1997 21st Century Software, New York City.
 *
 * See the file Reference/License.rtf for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by David Young <dwy@ace.net>, 1997.
 */

/* hacked by FMS and others to support NEXTSTEP */

#import "VNC.h"
#import "vncviewer.h"
#import "./NSByteOrder.h"

@implementation VNCFramebuffer

static inline unsigned RFBConvertPixel8ToPixel32(register unsigned char pixel8)
{
    register unsigned long pixel32 = 0;

#if defined(m68k) || defined(hppa) || defined(sparc) || defined(ppc)
    pixel32 += ((pixel8 >> 6) & 3) * 0x40000000;	// red
    pixel32 += ((pixel8 >> 4) & 3) * 0x00400000;	// green
    pixel32 += ((pixel8 >> 2) & 3) * 0x00004000;	// blue
#else
    pixel32 += ((pixel8 >> 6) & 3) * 0x00000040;	// red
    pixel32 += ((pixel8 >> 4) & 3) * 0x00004000;	// green
    pixel32 += ((pixel8 >> 2) & 3) * 0x00400000;	// blue
#endif

    return pixel32;
}

static inline unsigned RFBConvertPixel16ToPixel32(register unsigned short pixel16)
{
    register unsigned long pixel32 = 0;

#if defined(m68k) || defined(hppa) || defined(sparc) || defined(ppc)
    pixel32 += ((pixel16 >> 12) & 15) * 0x10000000;	// red
    pixel32 += ((pixel16 >>  8) & 15) * 0x00100000;	// green
    pixel32 += ((pixel16 >>  4) & 15) * 0x00001000;	// blue
#else
    pixel32 += ((pixel16 >> 12) & 15) * 0x00000010;	// red
    pixel32 += ((pixel16 >>  8) & 15) * 0x00001000;	// green
    pixel32 += ((pixel16 >>  4) & 15) * 0x00100000;	// blue
#endif

    return pixel32;
}


- (id)initWithSize:(NXSize)size
{
    if (self = [super init]) {
        _width = size.width;
        _height = size.height;
        _bytesPerRow = size.width * sizeof(unsigned);
		_minimumUpdateRect = NSMakeRect(0, 0, 0, 0);
        _framebufferCover = [[NSMutableData alloc] initWithLength:_bytesPerRow*_width];
        _framebufferPtr = [_framebufferCover mutableBytes];
    }

    return self;
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] %dx%d bytesPerRow=%d minUpdateRect=%%@ framebufferPtr=0x%lx>",
        [[self class] description], self,
        _width, _height, _bytesPerRow,
        _framebufferPtr];
}

- (void)dealloc
{
    [_framebufferCover release];

    [super dealloc];
}

- (NXImage *)framebufferImage
{
    NXSize size = {_width, _height};
    NXImage *image = [[[NXImage alloc] initSize:&size] autorelease];
    const unsigned char *pixelPointers[5];
    NXBitmapImageRep *imageRep;
    
    pixelPointers[0] = (const unsigned char *)_framebufferPtr;
    imageRep = [[[NXBitmapImageRep alloc]
        initDataPlanes:pixelPointers
                      pixelsWide:_width
                      pixelsHigh:_height
                   bitsPerSample:8
                 samplesPerPixel:3
                        hasAlpha:NO
                        isPlanar:NO
                  colorSpace:NX_RGBColorSpace
                     bytesPerRow:_bytesPerRow
                    bitsPerPixel:32] autorelease];

   	[image useRepresentation:imageRep];

    return image;
}

- (NSData *)subPixelsInRect:(NXRect)rect
{
    unsigned bytesPerRow = rect.size.width * sizeof(unsigned);
    NSMutableData *subData = [NSMutableData dataWithLength:bytesPerRow*rect.size.height];
    unsigned *subDataPtr = [subData mutableBytes];
    unsigned x, y;

    for (y = rect.origin.y; y < rect.origin.y+rect.size.height; ++y) {
	unsigned ystep=(y * _width);
        for (x = rect.origin.x; x < rect.origin.x+rect.size.width; ++x)
            *subDataPtr++ = _framebufferPtr[ystep+x];
    }

    return subData;
}

- (void)compositePixelData:(NSData *)pixelData targetRect:(NXRect)target bitsPerPixel:(unsigned)bitsPerPixel
{
    switch(bitsPerPixel) {
        case 32: {
            unsigned x, y;
            const unsigned *pixelPtr = [pixelData bytes];                          
            for (y = target.origin.y; y < (target.origin.y + target.size.height); ++y) {
		unsigned ystep=(y * _width);
                for (x = target.origin.x; x < (target.origin.x + target.size.width); ++x)
                    _framebufferPtr[ystep + x] = *pixelPtr++;
            }
            break;
        }

        case 16: {
            unsigned x, y;
            const unsigned short *pixelPtr = [pixelData bytes];                          
            for (y = target.origin.y; y < (target.origin.y + target.size.height); ++y) {
		unsigned int ystep=(y * _width);
                for (x = target.origin.x; x < (target.origin.x + target.size.width); ++x) {
                    _framebufferPtr[ystep + x] =  RFBConvertPixel16ToPixel32(*pixelPtr++);
   		}
	    }
            break;
	}
        case 8: {
           unsigned x, y;
           const unsigned char *pixelPtr = [pixelData bytes];                          
            for (y = target.origin.y; y < (target.origin.y + target.size.height); ++y) {
	       	unsigned int ystep=(y * _width);
                for (x = target.origin.x; x < (target.origin.x + target.size.width); ++x) {
                        _framebufferPtr[ystep + x] = RFBConvertPixel8ToPixel32(*pixelPtr++);
                 }
            }
            break;
 	}
        default:
            [NSException raise:NSInvalidArgumentException
                        format:@"%@ invalid bit depth %d",
                NSStringFromSelector(_cmd), bitsPerPixel];
    }

     //_minimumUpdateRect = NSUnionRect(_minimumUpdateRect, target);
    _minimumUpdateRect = *(NXUnionRect(&_minimumUpdateRect, &target));
}
- (void)compositeRawPixelData:(unsigned int *)pixelData targetRect:(NXRect)target bitsPerPixel:(unsigned)bitsPerPixel
{
    switch(bitsPerPixel) {
        case 32: {
            unsigned x, y;                         
            const unsigned *pixelPtr = pixelData;                          
            for (y = target.origin.y; y < (target.origin.y + target.size.height); ++y) {
		unsigned int ystep=(y * _width);
                for (x = target.origin.x; x < (target.origin.x + target.size.width); ++x)
                    _framebufferPtr[ystep + x] = *pixelPtr++; /* was NSSwapBigIntToHost(*pixelPtr++); */
            }
            break;
        }

        case 16: {
            unsigned x, y;
            const unsigned short *pixelPtr = ( const unsigned short *)pixelData;                          
            for (y = target.origin.y; y < (target.origin.y + target.size.height); ++y) {
		unsigned int ystep=(y * _width);
                for (x = target.origin.x; x < (target.origin.x + target.size.width); ++x) {
                    _framebufferPtr[ystep + x] =  RFBConvertPixel16ToPixel32(*pixelPtr++);
   		}
	    }
            break;
	}
        case 8: {
           unsigned x, y;
           const unsigned char *pixelPtr = (const unsigned char *) pixelData;                          
            for (y = target.origin.y; y < (target.origin.y + target.size.height); ++y) {
		unsigned int ystep=(y * _width);
                for (x = target.origin.x; x < (target.origin.x + target.size.width); ++x) {
                        _framebufferPtr[ystep + x] = RFBConvertPixel8ToPixel32(*pixelPtr++);
                 }
 	    }
            break;
 	}
        default:
            [NSException raise:NSInvalidArgumentException
                        format:@"%@ invalid bit depth %d",
                NSStringFromSelector(_cmd), bitsPerPixel];
    }
    _minimumUpdateRect = *(NXUnionRect(&_minimumUpdateRect, &target));
}

- (void)copyPixelsFromRect:(NXRect)srcRect toRect:(NXRect)destRect
{
    unsigned x, y;
    const unsigned *pixelPtr = [[self subPixelsInRect:srcRect] bytes];                          
    for (y = destRect.origin.y; y < (destRect.origin.y + destRect.size.height); ++y) {
	unsigned ystep=(y * _width);
      	for (x = destRect.origin.x; x < (destRect.origin.x + destRect.size.width); ++x)
         	_framebufferPtr[ystep + x] = *pixelPtr++;
    }
    _minimumUpdateRect = *(NXUnionRect(&_minimumUpdateRect, &destRect));
}

- (void)fillRect:(NXRect)rect color:(unsigned)color
{
    unsigned x, y;

    for (y = rect.origin.y; y < (rect.origin.y + rect.size.height); ++y)
        for (x = rect.origin.x; x < (rect.origin.x + rect.size.width); ++x)
            _framebufferPtr[(y * _width) + x] = color;
    _minimumUpdateRect = *(NXUnionRect(&_minimumUpdateRect, &rect));
}

- (void)fillRectList:(const NXRect *)rectList colorList:(const unsigned *)colorList count:(unsigned)count
{
    unsigned i = 0;

    while (i < count) {
        NXRect rect = rectList[i];
        unsigned x, y;
        unsigned color = colorList[i];

        for (y = rect.origin.y; y < (rect.origin.y + rect.size.height); ++y)
            for (x = rect.origin.x; x < (rect.origin.x + rect.size.width); ++x)
                _framebufferPtr[(y * _width) + x] = color;

    	_minimumUpdateRect = *(NXUnionRect(&_minimumUpdateRect, &rect));
        ++i;
    }
}

- (void)updateRect:(NXRect)updateRect
{
    const unsigned char *pixelPointers[5];
    NXBitmapImageRep *imageRep;

    /* balazs.pataki@sztaki.hu: this is a noop situation */
    if (updateRect.size.width == 0 || updateRect.size.height == 0) return;
	
    if (updateRect.size.width == _width && updateRect.size.height == _height)
        pixelPointers[0] = (const unsigned char *)_framebufferPtr;
    else
        pixelPointers[0] = [[self subPixelsInRect:updateRect] bytes];

    imageRep = [[NXBitmapImageRep alloc]
        initDataPlanes:pixelPointers
                      pixelsWide:updateRect.size.width
                      pixelsHigh:updateRect.size.height
                   bitsPerSample:8
                 samplesPerPixel:3
                        hasAlpha:NO
                        isPlanar:NO
                  colorSpace:NX_RGBColorSpace
                     bytesPerRow:updateRect.size.width*sizeof(unsigned int)
                    bitsPerPixel:32];

    // correct for coordinate weirdness
    updateRect.origin.y = (_height/2)+((_height/2)-updateRect.origin.y);
    updateRect.origin.y -= updateRect.size.height;
    
    [imageRep drawAt:&(updateRect.origin)];
    [imageRep free];
}

- (void)update
{
    [self updateRect:NSMakeRect(0, 0, _width, _height)];
    _minimumUpdateRect = NSMakeRect(0, 0, 0, 0);
}

- (void)updateMinimum
{
    [self updateRect:_minimumUpdateRect];
    _minimumUpdateRect = NSMakeRect(0, 0, 0, 0);
}

@end
